#include <mpi.h>
#define NUMBER_OF_BUFFERS 1

#define PARAMETER_TAG 63

int my_rank;
int comm_size;
MPI_Comm comm_comm;


typedef struct {
     int      tag;
     int      to_proc;
     int      from_proc;
     int      offset;
     int      indirect;
     int      data_type;
     int      length_index;
     int      length_int;
     int      length_double;
     int    * index;
     int    * int_val;
     double * double_val;
       } exchange_buffer ;

typedef struct {
     int               number_of_buffers;
     exchange_buffer * buffers;
     MPI_Request     * requests;
       } buffer_set;


typedef struct {
     int               length;
     MPI_Datatype      datatype;
     int             * proc;
     int             * temp;
     MPI_Request     * to_request_array;
     MPI_Status      * to_status_array;
     int             * to_buf_length;
     void           ** to_buf;
     MPI_Request     * from_request_array;
     MPI_Status      * from_status_array;
     int             * from_buf_length;
     void           ** from_buf;
       } exch_set_type;

void initialize_MPI(int *argc, char *** argv);
void finish_MPI();
void exit_MPI(int error );
void allocate_exchange_buffer(exchange_buffer * buffer , int length , int offset, int indirect, int data_type, int tag );
void deallocate_exchange_buffer(exchange_buffer * buffer);
void allocate_buffer_set(buffer_set * , int );
void fill_exchange_buffer(exchange_buffer buffer, double * double_values, int * int_values) ;
void initiate_send_buffer(double * , int *,  exchange_buffer, MPI_Request * );
void initiate_send_buffer_set( double * , int *, buffer_set );
void complete_send_buffer_set(buffer_set , MPI_Status * );
void initiate_receive_buffer(exchange_buffer , MPI_Request * );
void initiate_receive_buffer_set(buffer_set );
void apply_exchange_buffer(exchange_buffer , double * , int *);
void complete_receive_buffer_set(double * , int *, buffer_set );
void complete_sent_buffer(double * , int *, int , exchange_buffer * , MPI_Request * );

double dot_pro( double * , double * , int );

void allocate_exch_set(exch_set_type * exch , int length );
void initiate_send_exch_set(exch_set_type exch_set);
void initiate_receive_exch_set(exch_set_type exch_set);
void complete_receive_exch_set(exch_set_type exch_set);
void complete_send_exch_set(exch_set_type exch_set);
void sendreceive_exch_set(exch_set_type exch_set);
void define_neighbour_processors_exch_set(exch_set_type * exch ,int * array);
void hide_neighbour_processors_exch_set(exch_set_type * exch ,int * array);
void variables_exch_set(exch_set_type * exch, MPI_Datatype datatype,void * to_array, void *  from_array);
void arrays_exch_set(exch_set_type * exch, MPI_Datatype datatype
                        ,void * to_array,int * to_offset_array, int * to_length_array
                        ,void * from_array,int * from_offset_array, int * from_length_array
                        );
