#ifndef vector_DEF
   #define vector_DEF
   #include "vector.h"
#endif
#ifndef exchange_DEF
   #define exchange_DEF
   #include "exchange.h"
#endif
#ifndef timing_DEF
   #define timing_DEF
   #include "timing.h"
#endif
#include <stdlib.h>
#include <stdio.h>
#include <mpi.h>

/**************************************************************************************/
void alloc_vector_type(vector_type * aa, int length, int inner_length) {
 if ( length <  0 ) {
   printf("PE%i:\t error in routine alloc_vector_type:",my_rank);
   printf("PE%i:\t length=%d \n",my_rank, length);
   printf("PE%i:\t is less than 0!",my_rank);
       exit_MPI(1);
   exit(1);
                    };
   aa->length= length ;
   aa->inner_length= inner_length ;

if ( length > 0 ) {
   aa->values=malloc(sizeof(double)*length);
  } else {
   aa->values=NULL;
  }
}

/**************************************************************************************/
void read_vector_type(vector_type aa){
 
 if (aa.length<=0) {
   printf("PE%i:\t error in routine read_vector_type:",my_rank);
   printf("PE%i:\t length=%d \n",my_rank, aa.length);
   printf("PE%i:\t is less than 0!",my_rank);
       exit_MPI(1);
   exit(1);
                    };

   read_vector_type_kernel(aa.values, aa.length);
}
/**************************************************************************************/
void read_vector_type_kernel( double * aa, int length) {
        int i ;
        for (i=0; i<length; i++) {
               scanf( "%f",&aa[i]);
        }
}

/**************************************************************************************/
void write_vector_type(char string[], vector_type aa){

  printf("PE%i:\t \n",my_rank);
  printf("PE%i:\t %s\n",my_rank,string);
 
 if (aa.length< 0) {
   printf("PE%i:\t error in routine write_vector_type:\n",my_rank);
   printf("PE%i:\t length=%d \n",my_rank, aa.length);
   printf("PE%i:\t is less than 0!\n",my_rank);
       exit_MPI(1);
   exit(1);
                    };

   write_vector_type_kernel(aa.values, aa.length, aa.inner_length);
}
/**************************************************************************************/
void write_vector_type_kernel( double * aa, int length, int inner_length) {
        int i ;
               printf("PE%i:\t#inner part inner_length=%d  \n",my_rank,inner_length);
        for (i=0; i<inner_length; i++) {
               printf("PE%i:\t%d %f \n",my_rank,i,aa[i]);
        }
               printf("PE%i:\t#halo part length=%d  \n",my_rank,length-inner_length);
        for (i=inner_length; i<length; i++) {
               printf("PE%i:\t%d %f \n",my_rank,i,aa[i]);
        }
               printf("PE%i:\t \n",my_rank);
}


/**************************************************************************************/
void dealloc_vector_type(vector_type * vector){
 if ( vector->length > 0 ) {
   free(vector->values); vector->values=NULL; // NULL
 }
   vector->length=0;
}


/**************************************************************************************/
void set_vector(  vector_type aa ,int jmax, int kmax) {
      int ii, num,  jj, kk ;
      int num_row_elements ;

      ii=-1;
      num=-1;
   if   ( aa.length < (kmax-2)*(jmax-2) ) {
           printf("PE%i:\t error in routine set_vector:",my_rank);
           printf("PE%i:\t length= %d \n",my_rank,aa.length);
           printf("PE%i:\t is not enough long",my_rank);
       exit_MPI(1);
           exit(1);
   }
 set_vector_kernel(aa.values, aa.length, jmax, kmax); 
}
/**************************************************************************************/
void set_vector_kernel(double * aa, int length, int jmax, int kmax) {         
      int ii, num,  jj, kk ;
      int num_row_elements ;
      int jmin,kmin;

      ii=-1;
      num=-1;
      jmin=1;
      kmin=1;

           printf("PE%i:\t set_vector_kernel:",my_rank);
           printf("PE%i:\t  jmax=%d, kmax=%d \n",my_rank,jmax,kmax);
           printf("PE%i:\t  length=%d  \n",my_rank,length);
   for (kk=kmin; kk<=kmax; kk++){
     for( jj=jmin; jj<=jmax;jj++ ) {
      ii=ii+1;
      aa[ii]=1.;
      aa[ii]=(double)(jj-(jmin-1))/(double)((jmax+1)-(jmin-1)) ;
           printf("PE%i:\t  %d %f\n",my_rank,ii,aa[ii]);
     }
   }
}

/**************************************************************************************/
void vector_update( vector_type aa, vector_type bb, double  alpha ) {
  if   ( aa.length != bb.length) {
           printf("PE%i:\t error in routine vector_update:",my_rank);
           printf("PE%i:\t aa.length= %d ",my_rank, aa.length);
           printf("PE%i:\t is different from ",my_rank);
           printf("PE%i:\t bb.length= %d \n",my_rank, bb.length);
       exit_MPI(1);
           exit(1);
  }
   vector_update_kernel( aa.values, bb.values, alpha, aa.length) ;
}
/**************************************************************************************/
void vector_update_kernel( double * aa, double * bb, double  alpha, int length) {
        int i ;
	for (i=0; i<length; i++) {
		aa[i] = aa[i] + alpha * bb[i] ;
	}
}

/**************************************************************************************/
void vector_update_r(vector_type bb, vector_type aa, double alpha) {
    if (aa.length !=bb.length){
          printf("PE%i:\t error in routine vector_update_r:",my_rank);
          printf("PE%i:\t aa.length= %d ",my_rank, aa.length);
          printf("PE%i:\t is different from ",my_rank);
          printf("PE%i:\t bb.length= %d \n",my_rank, bb.length);
       exit_MPI(1);
          exit(1);
    }
  vector_update_r_kernel(bb.values, aa.values, alpha, aa.length);
}
/**************************************************************************************/
void vector_update_r_kernel( double * bb, double * aa, double  alpha, int length) {
        int i ;
	for (i=0; i<length; i++) {
		bb[i] = aa[i] + alpha * bb[i] ;
	}
}

/**************************************************************************************/
void vector_axpy( vector_type cc, vector_type aa, vector_type bb, double alpha){
     if (aa.length != bb.length){
          printf("PE%i:\t error in routine vector_axpy:",my_rank);
          printf("PE%i:\t aa.length= %d ",my_rank, aa.length);
          printf("PE%i:\t is different from ",my_rank);
          printf("PE%i:\t bb.length= %d \n",my_rank, bb.length);
       exit_MPI(1);
          exit(1);
     }
  vector_axpy_kernel(cc.values, aa.values, bb.values, alpha, aa.length);
}
/**************************************************************************************/
void vector_axpy_kernel(double * cc, double * aa, double * bb, double  alpha, int length) {
        int i ;
	for (i=0; i<length; i++) {
		cc[i] = aa[i] + alpha * bb[i] ;
	}
}

/**************************************************************************************/
void vector_to_vector(vector_type bb, vector_type aa){
  if (aa.length != bb.length){
          printf("PE%i:\t error in routine vector_to_vector:",my_rank);
          printf("PE%i:\t aa.length= %d ",my_rank, aa.length);
          printf("PE%i:\t is different from ",my_rank);
          printf("PE%i:\t bb.length= %d \n",my_rank, bb.length);
       exit_MPI(1);
          exit(1);
  }
   vector_to_vector_kernel(bb.values, aa.values, aa.length);
}
/**************************************************************************************/
void vector_to_vector_kernel(double * bb, double * aa, int length){
        int i;
        for (i=0; i<length; i++){
               bb[i] = aa[i];
        }
}


/**************************************************************************************/
void dotproduct( double * result, vector_type aa, vector_type bb){
     if (aa.length != bb.length){
          printf("PE%i:\t error in routine dotproduct:",my_rank);
          printf("PE%i:\t aa.length= %d ",my_rank, aa.length);
          printf("PE%i:\t is different from ",my_rank);
          printf("PE%i:\t bb.length= %d \n",my_rank, bb.length);
       exit_MPI(1);
          exit(1);
     }
  // length for vectors is aa.inner_length to exclude overlapping boundaries
//  dotproduct_kernel( result, aa.values, bb.values, aa.inner_length);
    *result=dot_pro( aa.values, bb.values, aa.inner_length);
}
/**************************************************************************************/
void dotproduct_kernel( double * result, double * aa, double * bb, int length ) {
	double sum=0;
        int i ;
	for (i=0; i<length; i++) {
	sum = sum + aa[i]*bb[i];
	}
	*result = sum;
}

/**************************************************************************************/
void exchange_vector(vector_type aa,buffer_set to_buffer_set,buffer_set from_buffer_set) {
   double time_initiate_send;
   double time_initiate_receive;
   double time_complete_receive;
   double time_complete_send;
   double time_start;
   double time_end;
   int * int_values;
      MPI_Status array_of_statuses[NUMBER_OF_BUFFERS];
   time_start=MPI_Wtime();

  initiate_send_buffer_set(aa.values,int_values, to_buffer_set);

//   time_initiate_send=MPI_Wtime()-time_start;

  initiate_receive_buffer_set(from_buffer_set);

 //  time_initiate_receive=MPI_Wtime()-time_start;

  complete_receive_buffer_set(aa.values,int_values,from_buffer_set);

 //  time_complete_receive=MPI_Wtime()-time_start;

  complete_send_buffer_set(to_buffer_set, array_of_statuses);

 //  time_complete_send=MPI_Wtime()-time_start;

 //  printf("PE%i:\t exchange_vector: time_initiate_send    %f muesec \n",my_rank, time_initiate_send);
 //  printf("PE%i:\t exchange_vector: time_initiate_receive %f muesec \n",my_rank, time_initiate_receive-time_initiate_send);
 //  printf("PE%i:\t exchange_vector: time_complete_receive %f muesec \n",my_rank, time_complete_receive-time_initiate_receive);
 //  printf("PE%i:\t exchange_vector: time_complete_send    %f muesec \n",my_rank, time_complete_send-time_complete_receive);
   printf("PE%i:\t exchange_vector: time %f muesec \n",my_rank,(time_end-time_start)*1.e6);

}
/**************************************************************************************/
void exch_vector(exch_set_type * exch_data, vector_type aa
             ,double * to_buffer_values,   int * to_buffer_index,   int * to_buffer_offset_array,   int * to_buffer_length_array
             ,double * from_buffer_values, int * from_buffer_index, int * from_buffer_offset_array, int * from_buffer_length_array
                              ) { 
   double time_complete_exchange;
   double time_start;
   int no;
   int offset;
   int ll;
   double help;
   int number_of_neighbours;

   time_start=MPI_Wtime();

number_of_neighbours=exch_data->length;

for( no=0; no < number_of_neighbours; no++){
   for( ll=0; ll < to_buffer_length_array[no]; ll++ ){
   offset=to_buffer_offset_array[no];
      to_buffer_values[offset+ll]=aa.values[to_buffer_index[offset+ll]];
//   printf("PE%i:\t exch_vector: to no= %d ll=%d offset=%d,to_buffer_index[offset+ll]=%d %f\n",my_rank
//                              ,no,ll,offset,to_buffer_index[offset+ll],to_buffer_values[offset+ll]);
   }
}

    arrays_exch_set( exch_data, MPI_DOUBLE
                        ,   to_buffer_values,   to_buffer_offset_array,   to_buffer_length_array
                        , from_buffer_values, from_buffer_offset_array, from_buffer_length_array
                        );

for( no=0; no < number_of_neighbours; no++){
   for( ll=0; ll < from_buffer_length_array[no]; ll++ ){
      offset=from_buffer_offset_array[no];
//   printf("PE%i:\t exch_vector: from no= %d ll=%d offset=%d,to_buffer_index[offset+ll]=%d %f\n",my_rank
//                             ,no,ll,offset,to_buffer_index[offset+ll],from_buffer_values[offset+ll]);
      aa.values[from_buffer_index[offset+ll]]=from_buffer_values[offset+ll];
   }
}


   time_complete_exchange=MPI_Wtime()-time_start;

   printf("PE%i:\t exch_vector: time_complete_exchange %f muesec \n",my_rank, time_complete_exchange*1.e6);

}
